# GUIDirectory.py
# $Id: GUIDirectory.py,v 1.27 2001/07/11 22:43:09 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
#
# Graphical representation of directory views.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
########################################################################

from gopher import *
from Tkinter import *
import Pmw
from string import *

import GopherResource
import GopherResponse
import GopherConnection
import ContentFrame
import Cache
import Options
import ResourceInformation

class GUIDirectory(ContentFrame.ContentFrame, Frame):
    # Behavior controllers
    verbose         = None

    # Colors for various links.
    DEFAULT_COLOR   = BLACK
    FOUND_COLOR     = RED
    LINK_COLOR      = BLUE
    ACTIVE_COLOR    = RED

    # Constants for use with packing things in their correct locations.
    TYPE_COLUMN     = 0
    TYPE_SPAN       = 1
    NAME_COLUMN     = 1
    NAME_SPAN       = 3
    BLURB_COLUMN    = 1
    BLURB_SPAN      = 3
    
    CACHE_COLUMN    = 4
    CACHE_SPAN      = 1
    INFO_COLUMN     = 5
    INFO_SPAN       = 2
    HOSTPORT_COLUMN = 7
    HOSTPORT_SPAN   = 1

    def __init__(self, parent_widget, parent_object, resp,
                 resource, filename=None, menuAssocs={}):
        Frame.__init__(self, parent_widget)  # Superclass constructor
        self.searchlist = []    # Searchable terms...
        self.parent = parent_object
        self.cursel = None
        self.resp   = resp
        self.resource   = resource
        self.menuAssocs = menuAssocs
        self.createPopup()

        pmsgbar = self.parent.getMessageBar()
        if pmsgbar:
            self.balloons = Pmw.Balloon(parent_widget,
                                        statuscommand=pmsgbar.helpmessage)
        else:
            self.balloons = Pmw.Balloon(parent_widget)

        if self.useStatusLabels:
            labeltext = "%s:%d" % (resource.getHost(), int(resource.getPort()))
        
            if resource.getName() != '' and resource.getLocator() != '':
                label2 = "\"%s\" ID %s" % (resource.getName(),
                                           resource.getLocator())
            else:
                label2 = "    "
            if len(label2) > 50:
                label2 = label2[0:47] + "..."
            
            Label(self, text=labeltext).pack(side='top', expand=0, fill='x')
            Label(self, text=label2).pack(side='top', expand=0, fill='x')

        self.scrolled_window = Pmw.ScrolledCanvas(self,
                                                  borderframe=1,
                                                  usehullsize=1,
                                                  hull_width=400,
                                                  hull_height=400,
                                                  hscrollmode='dynamic',
                                                  vscrollmode='dynamic')
        
        self.sbox = Frame(self.scrolled_window.interior())
        self.scrolled_window.create_window(0, 0, anchor='nw',
                                           window=self.sbox)
        self.sbox.bind('<Button-3>', self.framePopupMenu)
        self.scrolled_window.interior().bind('<Button-3>', self.framePopupMenu)

        # As very LAST thing, pack it in. 
        self.scrolled_window.pack(fill='both', expand=1)
        return None
        # End __init__

    def pack_content(self, *args):
        responses = self.resp.getResponses()

        def color_widget(event, self=self, *args):
            wid = event.widget
            wid.configure(foreground=self.ACTIVE_COLOR)
            return None

        # Rather than having to fetch a specified option each time through
        # the loop, fetch them all at the beginning, since they shouldn't
        # change while we're packing content into the window.
        tmpopts = Options.program_options
        info_in_directories = tmpopts.getOption("display_info_in_directories")
        show_host_port      = tmpopts.getOption("show_host_port")
        show_cached         = tmpopts.getOption('show_cached')

        for x in range(0, len(responses)):
            if x % 50 == 1:
                self.scrolled_window.resizescrollregion()
            
            r = responses[x]
            rname    = r.getName()
            rtype    = r.getType()
            
            l = Label(self.sbox, text=rtype)
            
            if r.getTypeCode() == RESPONSE_BLURB:
                # Some servers prefix blurbs that are not meant to be displayed
                # with (NULL).  Ferret these out, and just display a blank
                # line.
                if find(rname, "(NULL)") == 0:
                    rname = " "
                
                blurb_label = Label(self.sbox, foreground=self.DEFAULT_COLOR,
                                    text=rname)
                blurb_label.grid(row=x, columnspan=self.BLURB_SPAN,
                                 col=self.BLURB_COLUMN, sticky=W)
                self.searchlist.append([rname, blurb_label,
                                        self.DEFAULT_COLOR])
                
            else:
                # Trick Tk into passing arguments with the function and
                # get around Python's weird namespacing.
                def fn(event, self=self.parent, r=r, w=self, *args):
                    self.goElsewhere(r)
                    return None
                        
                def dopopup(event, resource=r, self=self):
                    # This binds resource to a parameter that
                    # one of several popup commands will use.
                    return self.popupMenu(event, resource)

                def enter_signal(event, resource=r,
                                 box=self.sbox, rowno=x, p=self):
                    host = resource.getHost()
                    port = resource.getPort()
                    resource.__blurb__ = Label(box,
                                               text=resource.toURL())
                    resource.__blurb__.grid(row=rowno,
                                            col=p.HOSTPORT_COLUMN,
                                            columnspan=p.HOSTPORT_SPAN,
                                            sticky=E)

                def leave_signal(event, resource=r):
                    try:
                        resource.__blurb__.destroy()
                    except:
                        pass
                
                # Don't make it clickable if it's an error.  But if it
                # isn't an error, connect these signals.
                if r.getTypeCode() != RESPONSE_ERR:
                    default_color = self.LINK_COLOR
                    b = Label(self.sbox, foreground=self.LINK_COLOR,
                              text=r.getName())
                    b.bind('<ButtonRelease-1>', fn)
                    b.bind('<Button-1>', color_widget)
                    b.bind('<Button-3>', dopopup)

                    if show_host_port:  # Cached option
                        b.bind('<Enter>', enter_signal)
                        b.bind('<Leave>', leave_signal)

                    # Attach a balloon
                    btext = r.toURL()
                    self.balloons.bind(b, None, btext)
                else:
                    default_color = self.DEFAULT_COLOR
                    b = Label(self.sbox, foreground=self.DEFAULT_COLOR,
                              text=rname)


                # Each entry in the searchlist is the name of a widget as a
                # string, the widget itself, and the widget's default color
                # The color is needed because 'finding' things requires
                # changing their color.
                self.searchlist.append([rname, b, default_color])
                
                l.grid(row=x,
                       col=self.TYPE_COLUMN,
                       columnspan=self.TYPE_SPAN,
                       sticky=W)
                b.grid(row=x,
                       col=self.NAME_COLUMN,
                       columnspan=self.NAME_SPAN,
                       sticky=W)

                cacheobj = self.parent.getCache()
                
                if info_in_directories:    # Cached option
                    if r.getInfo() != None:
                        i = r.getInfo()
                        t = i.getAdmin()
                        
                        if len(t) > 40:
                            t = t[0:40]
                            Label(self.sbox,
                                  text=t).grid(row=x,
                                               col=self.INFO_COLUMN,
                                               columnspan=self.INFO_SPAN)
                        else:
                            Label(self.sbox,
                                  text=t).grid(row=x,
                                               col=self.INFO_COLUMN,
                                               columnspan=self.INFO_SPAN)
                    else:
                        Label(self.sbox, text="  ").grid(row=x,
                                                    col=self.INFO_COLUMN,
                                                    columnspan=self.INFO_SPAN)

                # Possibly report to the user whether or not a given file is
                # present in cache.  I like to know this, but some people
                # don't, so make it a settable option.
                if show_cached:  # Cached option
                    if cacheobj.isInCache(r):
                        Label(self.sbox,
                              text="Cached").grid(row=x,
                                                  col=self.CACHE_COLUMN,
                                                  columnspan=self.CACHE_SPAN)
                    else:
                        Label(self.sbox,
                              text=" ").grid(row=x,
                                             col=self.CACHE_COLUMN,
                                             columnspan=self.CACHE_SPAN)

        self.scrolled_window.resizescrollregion()
        return None
        
    def destroy(self, *args):
        self.pack_forget()
        self.scrolled_window.destroy()
        Frame.destroy(self)
        return None

    def viewSource(self, *args):
        wintitle = "Source: %s" % self.resource.toURL()
        dialog   = Pmw.Dialog(self.parent, title=wintitle, buttons=())

        textwid = Pmw.ScrolledText(dialog.interior(), hscrollmode='dynamic',
                                   vscrollmode='static')
        textwid.insert('end', self.resp.toProtocolString())
        textwid.pack(expand=1, fill='both', side='top')
        return None

    def createPopup(self):
        """Pop-up menu on right click on a message"""
        self.popup = Menu(self)
        self.popup['tearoff'] = FALSE
        self.popup.add_command(label='Info', command=self.infoPopup)
        self.popup.add_command(label='Cache Status', command=self.cacheStatus)

        kz = self.menuAssocs.keys()
        kz.sort()
        
        for key in kz:
            self.popup.add_command(label=key,
                                   command=self.menuAssocs[key])

        self.framePopup = Menu(self)
        self.framePopup['tearoff'] = FALSE
        self.framePopup.add_command(label='View Directory Source',
                                    command=self.viewSource)

        for key in self.menuAssocs.keys():
            self.framePopup.add_command(label=key,
                                        command=self.menuAssocs[key])


    def popupMenu(self, event, item):
        """Display pop-up menu on right click on a message"""
        self.cursel = item
        self.popup.tk_popup(event.x_root, event.y_root)
        return None

    def framePopupMenu(self, event):
        self.framePopup.tk_popup(event.x_root, event.y_root)
        return None
    
    def cacheStatus(self, resource=None):
        c = self.parent.getCache()

        if resource is None:
            resource = self.cursel
            
        resp = c.isInCache(resource)

        if resp is None:
            str = "Resource\n%s\nis not in the cache." % resource.toURL()
            self.parent.genericMessage(str, "Cache Status")
        else:
            url = resource.toURL()
            [filename, size] = resp
            str = "%s\nis in the cache as\n%s\n and is %s bytes" % (url,
                                                                    filename,
                                                                    size)
            self.parent.genericMessage(str, "Cache Status")
        return None

    def infoPopup(self, resource=None):
        if resource is None:
            resource = self.cursel

        if not resource.isGopherPlusResource():
            # Info is only a valid operation if it's a Gopher+ resource.
            # Regular gopher servers don't provide any info in this way.
            str = "%s%s" % ("This item is not a Gopher+ Resource.\n",
                            "No special information is available about it.")
            self.parent.genericError(str, "Error:")
            return None
        
        if resource.getInfo() is None:
            try:
                conn = GopherConnection.GopherConnection()
                info = conn.getInfo(resource)
                resource.setInfo(info)
            except Exception, errstr:
                url = resource.toURL()
                str = "Cannot display information about\n%s:\n%s" % (url,
                                                                     errstr)
                self.parent.genericError(str, "Error:")
                return None

        info = resource.getInfo()
        str = "Resource information:\n%s\n%s" % (resource.toURL(),
                                                 info.toString())
        gri = ResourceInformation.GUIResourceInformation(info)
        return None
            
    def find(self, term, caseSensitive=None, lastIdentifier=None):
        """Overrides the same definition in ContentFrame.ContentFrame"""

        if lastIdentifier is None:  # Note this is a distinct case from when 
            self.lastIndex = -1     # lastIdentifier is false (possibly 0)
        else:
            # This could be 0 or any positive number.
            self.lastIndex = lastIdentifier

        try:
            self.lastIndex = int(self.lastIndex) # Unstringify
        except:
            raise(Exception,
                  "Illegal last ID passed to %s.find()" % self.__class__)

        if self.verbose:
            print "LAST FOUND INDEX was %s." % self.lastIndex
        
        # Bounds checking on lastIndex...
        if self.lastIndex < -1 or self.lastIndex > len(self.searchlist):
            print "*****Something's messed up.  Bad lastIdentifier."
        elif self.lastIndex >= 0:
            [old_label, old_widget, color] = self.searchlist[self.lastIndex]
            old_widget.configure(foreground=color)

        found      = None
        foundIndex = -1
        
        for x in range(self.lastIndex+1, len(self.searchlist)):
            [label, label_widget, color] = self.searchlist[x]
            
            if self.verbose:
                print "Looking at index %d through \"%s\"..." % (x, label)
                
            if not caseSensitive:
                # If we're not doing case-sensitive compares, convert them
                # both to lower case and we'll compare them that way.
                label = lower(label)
                term  = lower(term)
                
            if find(label, term) != -1:
                # Term was found in this label
                foundIndex = x
                found = 1
                # Find only one match per call to this function.  Bail out.
                break

        if found:
            [found_label, found_widget, color] = self.searchlist[foundIndex]

            # A bit of explanation on what this is:
            # to scroll to the correct location in the window (so the user
            # can see what was found and highlighted) we need a certain
            # distance to scroll the bar.  Since directory entries are all
            # that's in the window, and there's one per line, this should
            # work.
            scrlratio = float(foundIndex) / float(len(self.searchlist))

            if self.verbose:
                print "Scrlratio is: \"%s\"" % scrlratio
                print "Found \"%s\" in \"%s\" at index %d" % (term,
                                                              found_label,
                                                              foundIndex)
                
            # Turn the one that was found into FOUND_COLOR
            found_widget.configure(foreground=self.FOUND_COLOR)

            # Scroll the window to scrlratio - this should place the found
            # item at the very top of the window.
            self.scrolled_window.yview('moveto', scrlratio)
            
            if self.verbose:
                print "RETURNING NEW FOUND INDEX: %s" % foundIndex
                
            return foundIndex
        else:
            return None  # This signals that it wasn't found, and to reset

        # Everything else this class needs gets inherited from other classes.



# AssociationsEditor.py
# $Id: AssociationsEditor.py,v 1.8 2001/07/06 03:06:41 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
# This pops up a dialog box and allows the user to associate file name
# extensions with various programs to run.
#
# It returns an Associations object, and can optionally take one as an
# argument.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
########################################################################

import Associations
from gopher       import *
from Tkinter      import *
import Pmw
from string import *

class AssociationsEditor:
    DELIMITER = Associations.Associations.DELIMITER
    def __init__(self, parent, baseAssoc=Associations.Associations()):
        self.parent = parent
        self.dialog = Pmw.Dialog(parent, title='Edit Associations',
                                 buttons=('OK', 'Cancel'), defaultbutton='OK',
                                 command=self.dispatch)
        self.assoc = baseAssoc
        self.frame = self.dialog.interior()
        self.frame.pack(expand=1, fill='both')
        self.left_side  = Frame(self.frame)
        self.middle     = Frame(self.frame)
        self.right_side = Frame(self.frame)

        self.left_side.pack(side='left', expand=1, fill='both')
        self.middle.pack(side='left', expand=1, fill='both')
        self.right_side.pack(side='left', expand=1, fill='both')

        # Left side widgets
        inputbox = Frame(self.left_side)
        inputbox.grid(row=0, col=0, columnspan=2, rowspan=2, sticky=W)

        # File extension entry box and label.
        ftlabel = Label(inputbox, text='File extension:')
        self.ftEntry = Entry(inputbox, width=6)
        ftlabel.grid(row=0, col=0, sticky=W)
        self.ftEntry.grid(row=0, col=1, sticky=W)

        # Application entry box and label
        applabel = Label(inputbox, text='Application:')
        self.appEntry = Entry(inputbox, width=30)
        applabel.grid(row=1, col=0, sticky=W)
        self.appEntry.grid(row=1, col=1, sticky=W)

        # Instruction group box.
        group = Pmw.Group(self.left_side, tag_text='Instructions:')
        group.grid(row=2, column=0, rowspan=2, columnspan=2, sticky=W)
        instr1 = Label(group.interior(),
                       text='When entering in programs associated with files,')
        instr2 = Label(group.interior(),
                       text='Use $1 to represent the file being launched')
        instr3 = Label(group.interior(),
                       text='Filename extensions may be provided with or')
        instr4 = Label(group.interior(),
                       text='dots.  I.e. ".html" is the same as "html"')
        instr5 = Label(group.interior(),
                       text="Example: .html might be associated with")
        instr6 = Label(group.interior(),
                       text="netscape $1")
        instr7 = Label(group.interior(),
                       text=" ")
        instr8 = Label(group.interior(),
                       text="Extensions are case-sensitive")
        instr1.pack(side='top')
        instr2.pack(side='top')
        instr3.pack(side='top')
        instr4.pack(side='top')
        instr5.pack(side='top')
        instr6.pack(side='top')
        instr7.pack(side='top')
        instr8.pack(side='top')
        
        # Middle widgets
        self.addAssociationButton    = Button(self.middle, text='Add',
                                              command=self.add)
        self.removeAssociationButton = Button(self.middle, text='Remove',
                                              command=self.remove)
        self.setDefaultsButton       = Button(self.middle, text='Defaults',
                                              command=self.resetAssociations)
        # self.addAssociationButton.pack(side='top', expand=1, fill='both')
        # self.removeAssociationButton.pack(side='bottom', expand=1, fill='both')
        self.addAssociationButton.grid(row=0, col=0, sticky='NEW')
        self.removeAssociationButton.grid(row=1, col=0, sticky='NEW')
        self.setDefaultsButton.grid(row=2, col=0, sticky='NEW')
        
        # Right side widgets
        self.associationList = Pmw.ScrolledListBox(self.right_side,
                                                   hscrollmode='dynamic',
                                                   vscrollmode='static',
                                                   labelpos='nw',
                                                   dblclickcommand=self.reIns,
                                                   label_text='Associations:')
        self.associationList.pack(expand=1, fill='both')
        self.setAssociations(self.assoc)

        # self.dialog.activate()   # Make the dialog modal so the user can't
        #                          # mess with things in the other window.

    def resetAssociations(self, *args):
        self.setAssociations(self.assoc)

    def reIns(self, *args):
        selected = self.associationList.getcurselection()
        selected = selected[0]
        index = find(selected, self.DELIMITER)
        extension = selected[0:index]
        pgm = selected[index+len(self.DELIMITER):]

        self.ftEntry.delete(0, 'end')
        self.appEntry.delete(0, 'end')
        self.ftEntry.insert('end', extension)
        self.appEntry.insert('end', pgm)
        
        return None

    def extensionToAssociationExtension(self, ext):
        if len(ext) > 0:
            if ext[0] == '.':
                return ext[1:]
            else:
                return ext
        return ext
        
    def add(self, *args):
        extension = self.extensionToAssociationExtension(self.ftEntry.get())
        pgm       = self.appEntry.get()

        # Set the contents of the entry boxes back to nothing so the user
        # doesn't have to delete the contents before adding another association
        self.appEntry.delete(0, 'end')
        self.ftEntry.delete(0, 'end')

        str = extension + self.DELIMITER + pgm

        items = self.associationList.get()

        addItem = 1

        # Check to see if this entry is already in there somewhere.
        for x in range(0, len(items)):
            item = items[x]
            # If they have the same extension...
            if extension == item[0:len(extension)]:
                print "Replacing \"%s\"" % item
                # Remove it from the list.
                items = items[0:x-1] + (str,) + items[x+1:]
                addItem = None
                break

        if addItem:
            items = items + (str,)

        self.associationList.setlist(items)
        return None

    def remove(self, *args):
        self.associationList.delete('active')
        return None

    def setAssociations(self, assoc):
        list = ()
        
        for assocKey in assoc.getFileTypes():
            str = assocKey + self.DELIMITER + assoc.getProgramString(assocKey)
            list = list + (str,)

        self.associationList.setlist(list)
        return None

    def getAssociations(self, *args):
        self.assoc = Associations.Associations()

        for item in self.associationList.get():
            print "Got item %s" % item
            index = find(item, self.DELIMITER)
            extension = item[0:index]
            pgm = item[index+len(self.DELIMITER):]
            self.assoc.addAssociation(extension, pgm)
        
        return self.assoc
    
    def dispatch(self, button):
        if button == 'OK':
            assocs = self.getAssociations()
            self.parent.setAssociations(assocs)
            self.dialog.destroy()
            # Grab data and put it into an Assocations object.

        self.dialog.destroy()
        return None




# ResourceInformation.py
# $Id: ResourceInformation.py,v 1.6 2001/07/05 17:16:53 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
# Released under the terms of the GNU General Public License
#
# When dealing with a Gopher+ server, information about a document can be
# fetched by sending the request:
# some_locator\t!\r\n
#
# This module handles the parsing and storing of that information.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
############################################################################

from string import *
from Tkinter import *
import Pmw
import os
import re
import ContentFrame
from gopher         import *
import GopherResource
import GopherResponse

class ResourceInformation:
    verbose = None
    def __init__(self, data=None):
        self.blockdict = {}

        self.data = data
        
        if self.data != None:
            self.setData(self.data)

        return None
    def __str__(self):
        return self.toString()
    def toString(self):
        def fn(key, obj=self):
            return "%s:\n%s\n" % (upper(key), obj.getBlock(key))
        
        return join(map(fn, self.getBlockNames()), "") + "\n"
    
    def setData(self, data):
        self.data = data
        self.data = re.sub("\r\n", "\n", self.data)

        lastindex = -1
        blocks = []

        try:
            while 1:
                # This will throw ValueError if not found.
                newindex = index(self.data, "\n+",
                                 (lastindex + 1), len(self.data))
                blocks.append(self.data[lastindex+1:newindex])
                lastindex = newindex
        except ValueError:  # When no more "\n+" are found.
            # The remaining block is what's left...
            blocks.append(self.data[lastindex+1:len(self.data)])

        # What we're going to do is hash each data block in by its block
        # 'title'.  This way, when we're done hashing all of the data, we
        # can just go through and pick out what we want.  So this type of
        # data:
        # +ADMIN:
        #  Grendel The Evil <grendel@nowhere.com>
        #  Some more admin information
        # Gets hashed in like this essentially:
        # hash['admin'] = "Grendel The Evil <grendel@nowhere.com>\n..."
        self.blockdict = {}

        # We now have a list of blocks.
        for block in blocks:
            lines = split(block, "\n")
            blocklabel = lines[0]

            front = find(blocklabel, "+")   # This defines the start of a block
            back  = find(blocklabel, ":")   # This may not be present

            if front != -1:
                if back == -1:
                    back = len(blocklabel)  # End of string if not present

                # Get the name, which is like this: "+ADMIN:" => 'ADMIN'
                blockname = blocklabel[front+1:back]
                key = lower(blockname) # Lowercase so it hashes nicely.  :)
                
                # strip the leading space.  This is because in gopher+
                # when it responds to info queries, each response line that
                # isn't a block header is indented by one space.
                data = re.sub("\n ", "\n", join(lines[1:], "\n"))

                # Get the first space in the data.
                if self.verbose:
                    print "Data is %s" % data

                # Watch out for the case when data is ''.  This could be in
                # particular if the server sends us a size packet like this:
                # +-1\r\n
                # Which would have a '' data segment.
                if data != '' and data[0] == ' ':
                    data = data[1:]

                # Assign it into the hash.
                if self.verbose:
                    print "Assigned data to key %s" % key

                if data != '' and not data is None:
                    # No sense in assigning nothing into a key.  The getBlock()
                    # handles when there is no data and returns ''
                    self.blockdict[key] = data
            else:
                print "BLOCK ERROR: cannot find blockname in %s" % blocklabel

        if self.verbose:
            k = self.blockdict.keys()
            print "Available block titles are:\n%s" % join(k, "\n")

        print "Keys are ", self.blockdict.keys()
        return self

    # Simple accessors/mutators.
    # Sure, I could just monkey around with the data in an object from outside,
    # but in some cultures, people are executed for such offenses against the
    # OOP paradigm.  :)
    def setBlock(self, blockname, blockval):
        self.blockdict[lower(blockname)] = blockval
        return self.getBlock(lower(blockname))
    def setInfo(self, newinfo):
        self.blockdict['info'] = newinfo
        return self.getInfo()
    def setAdmin(self, newadmin):
        self.blockdict['admin'] = newadmin
        return self.getAdmin()
    def setViews(self, newviews):
        self.blockdict['views'] = newviews
        return self.getViews()
    def setAbstract(self, newabstract):
        self.blockdict['abstract'] = newabstract
        return self.getAbstract()
    def getAbstract(self):
        return self.blockdict['abstract']
    def getViews(self):
        return self.blockdict['views']
    def getInfo(self):
        return self.blockdict['info']
    def getAdmin(self):
        return self.blockdict['admin']
    def getBlockNames(self):
        return self.blockdict.keys()
    def getBlock(self, blockname):
        try:
            return self.blockdict[lower(blockname)]
        except KeyError:
            return ''

class GUIResourceInformation(Pmw.TextDialog):
    def __init__(self, resource_info_object):
        Pmw.TextDialog.__init__(self, title="Resource Information")
        self.insert('end', resource_info_object.toString())
        return None
    
        



# GopherConnection.py
# $Id: GopherConnection.py,v 1.16 2001/07/14 22:51:57 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
# Released under the terms of the GNU General Public License
#
# This object handles connections and sending requests to gopher servers.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#######################################################################
import socket
import re
import utils
from string import *
import GopherResponse
import Connection
import GopherObject
import GopherResource
import ResourceInformation
import AskForm
from gopher import *

GopherConnectionException = "Error: "

class GopherConnection(Connection.Connection):
    SNARFSIZE                 = 1024
    verbose                   = None
    def __init__(self, server="gopher://gopher.floodgap.com", port=70):
        Connection.Connection.__init__(self)
        self.server = re.sub("gopher://", "", server, 1)
        self.port   = port
        self.forgetResponse()
        return None

    def forgetResponse(self):
        """Remove the response field of this object.  This is periodically
        necessary, since it can get quite large."""
        self.response = None
        return self.response
    def stripTail(self, data):
        ind = rfind(data, "\r\n.\r\n")
        
        if ind != -1:
            if self.verbose:
                print "Stripping protocol footer at index %d" % int(ind)
            return data[0:ind]

        if self.verbose:
            print "No protocol footer found."
        return data

    # Get extended information about a resource.
    def getInfo(self, resource, msgBar=None):
        try:
            req = "%s\t!\r\n" % resource.getLocator()
            data = self.requestToData(resource, req, msgBar)
        except Connection.ConnectionException, errstr:
            raise GopherConnectionException, errstr
        
        if self.verbose:
            print "Got %d bytes from INFO conn:\n%s" % (len(data),
                                                        data)

        # The server sends back a length of -2 when it doesn't know how long
        # the document is, and when the document may contain the \r\n.\r\n
        # pattern.  So if it might, do not strip it out.  Otherwise do.  This
        # will probably keep out a very subtle bug of having files downloaded
        # be truncated in really weird places.  (Where the \r\n.\r\n would have
        # been)
        if resource.getLen() != -2:
            if self.verbose:
                print "Stripping protocol footer."
            data = self.stripTail(data)
            
        try:
            info = ResourceInformation.ResourceInformation(data)
        except Exception, estr:
            print "*********************************************************"
            print "***GopherConnection: ResourceInformation Error: %s" % estr
            print "*********************************************************"
            raise GopherConnectionException, estr
        
        return info
            
    def getResource(self, resource, msgBar=None):
        self.forgetResponse()
        self.host     = re.sub("gopher://", "", resource.getHost(), 1)
        self.port     = resource.getPort()
        self.lastType = resource.getType()

        try:
            if resource.getDataBlock():
                request = resource.getLocator() + "\t+\t1\r\n"
                request = request + resource.getDataBlock()
                
                self.response = self.requestToData(resource,
                                                   request,
                                                   msgBar, 1)
            elif resource.isGopherPlusResource() and resource.isAskType():
                info = resource.getInfo(shouldFetch=1)
                af = AskForm.AskForm(info.getBlock("ASK"))

                # Copy host/port/locator information into af
                af.dup(resource)
                return af
            elif resource.isGopherPlusResource():
                request = resource.getLocator() + "\t+\r\n"
                self.response = self.requestToData(resource,
                                                   request,
                                                   msgBar, 1)
            else:
                request = resource.getLocator() + "\r\n"
                self.response = self.requestToData(resource,
                                                   request,
                                                   msgBar, None)
        except Connection.ConnectionException, estr:
            error_resp = GopherResponse.GopherResponse()
            errstr = "Cannot fetch\n%s:\n%s" % (resource.toURL(),
                                                estr)

            error_resp.setError(errstr)
            return error_resp

        utils.msg(msgBar, "Examining response...")

        resp = GopherResponse.GopherResponse()
        resp.setType(resource.getTypeCode())

        if resource.getLen() != -2:
            self.response = self.stripTail(self.response)

        try:
            # The parser picks up directory entries and sets the internal
            # data of the object as needed.
            resp.parseResponse(self.response)

            # if we get this far, then it's a directory entry, so set the
            # data to nothing.
            resp.setData(None)
        except Exception, erstr:
            # print "Caught exception while parsing response: \"%s\"" % erstr
            if self.verbose:
                print "OK, it's data."
            resp.setData(self.response)
                
        return resp



# State.py
# $Id: State.py,v 1.2 2001/07/11 22:43:09 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
#
# Saves state information about a particular GopherResource.
# This is a bit overkill, but it was easier than putting the information
# in a tuple and then having to worry about which index gets to which item
# in the tuple.  Easier just to call methods like getResource()  :)
#
# This is just a souped up structure.  No calculation is done, just accessors
# and mutators are provided to bundle several different data items together
# under one object.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
############################################################################
class State:
    verbose = None
    
    def __init__(self, response, resource, widget):
        self.response = response
        self.resource = resource
        self.widget   = widget
        return None
    
    def __str__(self):
        return "%s" % self.resource.toURL()
    
    def __repr__(self):
        return self.__str__()

    # Accessors
    
    def getResponse(self):
        return self.response
    
    def getResource(self):
        return self.resource
    
    def getWidget(self):
        return self.widget

    # Mutators
    
    def setResponse(self, resp):
        self.response = resp
        return self.getResponse()
    
    def setResource(self, res):
        self.resource = res
        return self.getResource()
    
    def setWidget(self, widget):
        self.widget = widget
        return self.getWidget()



# Tree.py
# highly optimized tkinter tree control
# Written by Charles E. "Gene" Cash <gcash@fdn.com>
# Modifications by David Allen <mda@idatar.com>
#
# 98/12/02 CEC started
# 99/??/?? CEC release to comp.lang.python.announce
#
# to do:
# add working "busy" cursor
# make paste more general
# add good ideas from win95's tree common control
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
############################################################################
            
import os
import string
import Tkdnd
import Bookmark
import ListNode
import List
import BookmarkEditor
import Dialogs
from Tkinter import *

# this is initialized later, after Tkinter is started
class Icons:
    OPEN_ICON = None
    SHUT_ICON = None
    FILE_ICON = None

#------------------------------------------------------------------------------
# cut'n'paste helper object

class Cut_Object:
    def __init__(self, node=None, id=None, state=None):
        self.node    = node
        self.full_id = id
        self.state   = state
    def getNode(self):
        return node
    def getFullId(self):
        return self.full_id
    def getState(self):
        return state
        
#------------------------------------------------------------------------------
# tree node helper class

class NodeData:
    """This is the same as Node, but only holds the data and does nothing with
    it.  This is ideal for creating nodes that you don't want drawn until they
    get pasted in via a Cut_Object"""
    def __init__(self, parent, name, id, closed_icon, open_icon, x, y,
                 parentwidget):
        self.parent      = parent
        self.name        = name
        self.id          = id
        self.open_icon   = open_icon
        self.closed_icon = closed_icon
        self.widget      = parentwidget
        self.subnodes    = []
        self.spinlock    = 0
        self.openflag    = 0
        return None
    
class Node:
    # initialization creates node, draws it, and binds mouseclicks
    def __init__(self, parent, name, id, closed_icon, open_icon, x, y,
                 parentwidget):
        # lots of things to remember
        # immediate parent node
        self.parent = parent

        self.__x = x
        self.__y = y
        
        # name displayed on the label
        self.name = name
        
        # internal name used to manipulate things
        self.id = id
        
        # bitmaps to be displayed
        self.open_icon   = open_icon
        self.closed_icon = closed_icon
        
        # tree widget we belong to
        self.widget = parentwidget
        
        # our list of child nodes
        self.subnodes = []
        
        # cheap mutex spinlock
        self.spinlock = 0
        # closed to start with
        self.open_flag = 0
            
        # call customization hook
        if self.widget.init_hook:
            self.widget.init_hook(self)

    def drawNode(self):
        # draw horizontal connecting lines
        if self.widget.lineflag:
            self.line = self.widget.create_line(self.__x-self.widget.distx,
                                                self.__y, self.__x, self.__y)

        # draw approprate image
        self.symbol = self.widget.create_image(self.__x, self.__y,
                                               image=self.closed_icon)

        # add label
        self.label = self.widget.create_text(self.__x+self.widget.textoff,
                                             self.__y, 
                                             text=self.name, justify='left',
                                             anchor='w')

        def infoCallback(event, s=self):
            display_information(s, s.widget)
        
        self.widget.tag_bind(self.label, '<Double-Button-1>', infoCallback)

        # single-click to expand/collapse
        cmd = self.widget.tag_bind(self.symbol, '<1>', self.click)
            
        # drag'n'drop support
        if self.widget.dnd_hook:
            # this starts the drag operation
            self.widget.tag_bind(self.label, '<1>',
                                 lambda ev,
                                 qs=self:qs.widget.dnd_click(ev, qs))

            # these events help identify drop target node
            self.widget.tag_bind(self.symbol, '<Enter>',
                                 lambda ev, qs=self:qs.widget.enter(qs))
            self.widget.tag_bind(self.label, '<Enter>',
                                 lambda ev, qs=self:qs.widget.enter(qs))
            self.widget.tag_bind(self.symbol, '<Leave>', self.widget.leave)
            self.widget.tag_bind(self.label, '<Leave>', self.widget.leave)
        elif self.widget.item_dbl_click_hook:
            self.widget.tag_bind(self.label, '<Double-Button-1>',
                                 self.widget.item_dbl_click_hook)

    def isFile(self):
        return (self.children() == 0 and
                (self.open_icon == None or self.open_icon == Icons.FILE_ICON))

    def isFolder(self):
        return not self.isFile()

    def children(self):
        return len(self.subnodes)

    def __repr__(self):
        try:
            # Parent may be None so this may croak.
            pname = self.parent.name 
        except:
            pname = "NONEXISTANT"
            
        return 'Node: %s  Parent: %s  (%d children)' % \
               (self.name, pname, len(self.subnodes))
    
    # recursively delete subtree & clean up cyclic references
    def _delete(self):
        for i in self.subnodes:
            if i.open_flag and i.subnodes:
                # delete vertical connecting line
                if self.widget.lineflag:
                    self.widget.delete(i.tree)

            # delete node's subtree, if any
            i._delete()

            # the following unbinding hassle is because tkinter
            # keeps a callback reference for each binding
            # so if we want things GC'd...
            #
            # MDA: 7/5/2001: Commented this out because it breaks with
            # Python 2.1's default Tkinter as _tagcommands doesn't exist.
            # Followup.
            # for j in (i.symbol, i.label):
            #    for k in self.widget.tag_bind(j):
            #        self.widget.tag_unbind(j, k)
            #    for k in self.widget._tagcommands.get(j, []):
            #        self.widget.deletecommand(k)
            #        self.widget._tagcommands[j].remove(k)
                    
            # delete widgets from canvas
            self.widget.delete(i.symbol, i.label)
            if self.widget.lineflag:
                self.widget.delete(i.line)
                
            # break cyclic reference
            i.parent=None

        # move cursor if it's in deleted subtree
        if self.widget.pos in self.subnodes:
            self.widget.move_cursor(self)
            
        # now subnodes will be properly garbage collected
        self.subnodes = []

    # move everything below current icon, to make room for subtree
    # using the magic of item tags
    def _tagmove(self, dist):
        # mark everything below current node as movable
        bbox1 = self.widget.bbox(self.widget.root.symbol, self.label)
        bbox2 = self.widget.bbox('all')
        self.widget.dtag('move')
        self.widget.addtag('move', 'overlapping', 
                           bbox2[0], bbox1[3], bbox2[2], bbox2[3])

        # untag cursor & node so they don't get moved too
        # this has to be done under Tk on X11
        self.widget.dtag(self.widget.cursor_box, 'move')
        self.widget.dtag(self.symbol, 'move')
        self.widget.dtag(self.label, 'move')

        # now do the move of all the tagged objects
        self.widget.move('move', 0, dist)

        # fix up connecting lines
        if self.widget.lineflag:
            n = self
            while n:
                if len(n.subnodes) and n.subnodes[-1] is not None:
                    # position of current icon
                    x1, y1 = self.widget.coords(n.symbol)
                    # position of last node in subtree
                    x2, y2 = self.widget.coords(n.subnodes[-1:][0].symbol)
                    self.widget.coords(n.tree, x1, y1, x1, y2)
                n = n.parent

    # return list of subnodes that are expanded (not including self)
    # only includes unique leaf nodes (e.g. /home and /home/root won't
    # both be included) so expand() doesn't get called unnecessarily
    # thank $DEITY for Dr. Dutton's Data Structures classes at UCF!
    def expanded(self):
        # push initial node into stack
        stack = [(self, (self.id,))]
        list = []
        while stack:
            # pop from stack
            p, i = stack[-1:][0]
            del stack[-1:]
            # flag to discard non-unique sub paths
            flag = 1
            # check all children
            for n in p.subnodes:
                # if expanded, push onto stack
                if n.open_flag:
                    flag = 0
                    stack.append(n, i+(n.id,))
            # if we reached end of path, add to list
            if flag:
                list.append(i[1:])
        return list

    # get full name, including names of all parents
    def full_id(self):
        if self.parent:
            return self.parent.full_id()+(self.id,)
        else:
            return (self.id,)

    # expanding/collapsing folders
    def toggle_state(self, state=None):
        if not self.open_icon:
            # not a expandable folder
            return
        if state == None:
            # toggle to other state
            state = not self.open_flag
        else:
            # are we already in the state we want to be?
            if (not state) == (not self.open_flag):
                return

        # not re-entrant
        # acquire mutex
        while self.spinlock:
            pass
        self.spinlock = 1
        
        # call customization hook
        if self.widget.before_hook:
            self.widget.before_hook(self)

        # if we're closed, expand & draw our subtrees
        if not self.open_flag:
            self.open_flag = 1
            self.widget.itemconfig(self.symbol, image=self.open_icon)

            # get contents of subdirectory or whatever
            contents = self.widget.get_contents(self)

            # move stuff to make room
            self._tagmove(self.widget.disty*len(contents))

            # now draw subtree
            self.subnodes = []
            
            # get current position of icon
            x, y = self.widget.coords(self.symbol)
            yp = y
            
            for i in contents:
                try:
                    print "CONTENTS: i is %s class %s" % (i, i.__class__)
                except:
                    pass
                # add new subnodes, they'll draw themselves
                yp = yp+self.widget.disty
                newnode = Node(self, i[0], i[1], i[2], i[3],
                               x+self.widget.distx, yp,
                               self.widget)
                newnode.drawNode()
                self.subnodes.append(newnode)
                
            # the vertical line spanning the subtree
            if self.subnodes and self.widget.lineflag:
                _tval = y+self.widget.disty*len(self.subnodes)
                self.tree = self.widget.create_line(x, y, x, _tval)
                self.widget.lower(self.tree, self.symbol)

        # if we're open, collapse and delete subtrees
        elif self.open_flag:
            self.open_flag = 0
            self.widget.itemconfig(self.symbol, image=self.closed_icon)

            # if we have any children
            if self.subnodes:
                # recursively delete subtree icons
                self._delete()

                # delete vertical line
                if self.widget.lineflag:
                    self.widget.delete(self.tree)

                # find next (vertically-speaking) node
                n = self
                while n.parent:
                    # position of next sibling in parent's list
                    i = n.parent.subnodes.index(n)+1
                    if i < len(n.parent.subnodes):
                        n = n.parent.subnodes[i]
                        break
                    n = n.parent
                    
                if n.parent:
                    # move everything up so that distance to next subnode is
                    # correct
                    x1, y1 = self.widget.coords(self.symbol)
                    x2, y2 = self.widget.coords(n.symbol)
                    dist = y2-y1-self.widget.disty
                    self._tagmove(-dist)

        # update scroll region for new size
        x1, y1, x2, y2 = self.widget.bbox('all')
        self.widget.configure(scrollregion=(x1, y1, x2+5, y2+5))

        # call customization hook
        if self.widget.after_hook:
            self.widget.after_hook(self)

        # release mutex
        self.spinlock = 0

    def expandAll(self):
        self.toggle_state(1) # Expand this item
        
        for n in self.subnodes:
            # Recursively expand subnodes.
            n.expandAll()
            
        return None

    # expand this subnode
    # doesn't have to exist, it expands what part of the path DOES exist
    def expand(self, dirs):
        # if collapsed, then expand
        self.toggle_state(1)

        # find next subnode
        if dirs:
            for n in self.subnodes:
                if n.id == dirs[0]:
                    n.expand(dirs[1:])
                    break
    
    # handle mouse clicks by moving cursor and toggling folder state
    def click(self, event):
        self.widget.move_cursor(self)
        self.toggle_state()

    # cut a node and it's subtree
    def cut(self):
        # remember what was expanded, so we can re-expand on paste
        expand_list = self.expanded()
        if not self.open_flag:
            expand_list = None

        id = self.full_id()
        
        # collapse
        self.toggle_state(0)

        # delete from tree
        if self.parent:
            # Remove all data from the parent's BookmarkMenu
            # print "Removing reference from \"%s\" to \"%s\"" % (
            #     self.parent.id, self.id)
            # print "Class of removee is ", self.id.__class__
            self.parent.id.removeReference(self.id)
            
            # move cursor safely out of the way
            if self.widget.pos == self:
                self.widget.prev()
                
            if len(self.parent.subnodes) == 1:
                # delete vertical connecting line
                # if we're the only child
                if self.widget.lineflag:
                    self.widget.delete(self.parent.tree)

            # delete from parent's list of children
            self.parent.subnodes.remove(self)

        # move rest of tree up
        self._tagmove(-self.widget.disty)

        # break cyclic reference
        self.parent = None

        # see _delete() for why we have to do this
        # MDA: 7/5/2001: Commented this out because it breaks with
        # Python 2.1's default Tkinter as _tagcommands doesn't exist.
        # Followup.
        # for j in (self.symbol, self.label):
        #    for k in self.widget.tag_bind(j):
        #         self.widget.tag_unbind(j, k)
        #     for k in self.widget._tagcommands.get(j, []):
        #         self.widget.deletecommand(k)
        #         self.widget._tagcommands[j].remove(k)
                
        # delete from canvas
        self.widget.delete(self.symbol, self.label)
        if self.widget.lineflag:
            self.widget.delete(self.line)
            
        # update scrollbar for new height
        x1, y1, x2, y2 = self.widget.bbox('all')
        self.widget.configure(scrollregion=(x1, y1, x2+5, y2+5))

        # returns a "cut_object"
        co = Cut_Object(self, id, expand_list)
        
        # call customization hook
        if self.widget.cut_hook:
            self.widget.cut_hook(co)

        return co

    # insert a "cut object" at the proper place
    # option:
    # 1 - insert as 1st child
    # 2 - insert after last child
    # 3 - insert as next sibling
    def paste(self, co, option=1):
        # call customization hook
        # this usually does the actual cut'n'paste on the underlying
        # data structure
        if self.widget.paste_hook:
            # if it returns false, it wasn't successful
            if self.widget.paste_hook(co):
                return

        self.expand([1])
        # expand if necessary
        # if option == 1 or option == 2:
        #     self.toggle_state(2)

        # Uncomment if lineflag to Tree() was true.
        #if option == 1 or option == 2 and not self.subnodes:
        #    # Create the horizontal line if it isn't already present.
        #    # (i.e. inserting into a dir that has 0 children)
        #    # Warning: this is an ugly hack.
        #    x, y = self.widget.coords(self.symbol)
        #    _tval = y+self.widget.disty*len(self.subnodes)
        #    self.tree = self.widget.create_line(x, y, x, _tval)
        #    self.widget.lower(self.tree, self.symbol)
            
        # make subnode list the right size for _tagmove()
        if option == 1 or not self.parent:
            self.subnodes.insert(0, None)
            # Option is insert as first child, so prepend the Bookmark
            # to the BookmarkMenu
            self.id.prepend(ListNode.ListNode(co.node.id))
            i = 0
        elif option == 2:
            self.id.postpend(ListNode.ListNode(co.node.id))
            self.subnodes.append(None)
            i = -1
        elif option == 3:
            i = self.parent.subnodes.index(self)+1

            # Update the BookmarkMenu relationship
            self.parent.id.insertAfter(newnode=ListNode.ListNode(co.node.id),
                                       afterWhat=self.id)

            # Update the list of GUI subnodes
            self.parent.subnodes.insert(i, None)
            
        # move rest of tree down
        self._tagmove(self.widget.disty)

        # place new node
        xval, yval = self.widget.coords(self.symbol)
        if option == 1 or option == 2:
            xval = xval+self.widget.distx
        yval = yval+self.widget.disty

        # create node
        if option == 1 or option == 2:
            node_parent = self
        elif option == 3:
            node_parent = self.parent

        n = Node(parent=node_parent,
                 name=co.node.name,
                 id=co.node.id,
                 # co.node.open_flag,
                 # co.node.icon,
                 closed_icon=co.node.closed_icon,
                 open_icon=co.node.open_icon,
                 x=xval,
                 y=yval,
                 parentwidget=self.widget)
        n.drawNode()
        
        # insert into tree...if it's the root item, don't try to insert it as
        # a sibling.
        if option == 1 or option == 2 or not self.parent:
            self.subnodes[i] = n
        elif option == 3:
            # insert as next sibling
            self.parent.subnodes[i] = n
        
        # expand children to the same state as original cut tree
        if co.state:
            n.expand(co.state)

    # return next lower visible node
    def next(self):
        n = self
        if n.subnodes:
            # if you can go right, do so
            return n.subnodes[0]
        while n.parent:
            # move to next sibling
            i = n.parent.subnodes.index(n)+1
            if i < len(n.parent.subnodes):
                return n.parent.subnodes[i]
            # if no siblings, move to parent's sibling
            n = n.parent
        # we're at bottom
        return self
    
    # return next higher visible node
    def prev(self):
        n = self
        if n.parent:
            # move to previous sibling
            i = n.parent.subnodes.index(n)-1
            if i >= 0:
                # move to last child
                n = n.parent.subnodes[i]
                while n.subnodes:
                    n = n.subnodes[-1]
            else:
                # punt if there's no previous sibling
                if n.parent:
                    n = n.parent
        return n

def display_information(node, parent_widget):
    """Displays information about the given node in a separate window that
    is a child of parent_widget.  This is for when the user requests
    information about different tree items."""
    data = node.id

    # Since python does everything call by reference, we can create a callback
    # function that merely edits the data of the node in place, and it will
    # apply to what's actually in the tree.
    # Pass this the original reference (in the tree) and the new resource
    # whose information matches what the stuff in the tree should be.
    def editReference(orig_reference, new_res, n=node):
        if orig_reference.__class__ == Bookmark.Bookmark:
            orig_reference.setURL(new_res.getURL())
            orig_reference.setName(new_res.getName())
            # Update the label on the widget
            n.widget.itemconfig(n.label, text=new_res.getName())
        else:
            # It's a folder.
            orig_reference.setName(new_res)
            n.widget.itemconfig(n.label, text=new_res)

        return None

    # Ugly.  NewBookmarkDialog takes a callback, and this is it.  It will
    # pass the res argument - a newly created GopherResource object.
    # Call editReference on it instead with the required arguments.
    def editReferenceStub(res, ref=data, callback=editReference):
        return callback(orig_reference=ref,
                        new_res=res)

    if data.__class__ == Bookmark.Bookmark:
        # Create a new bookmark dialog.  Note this is the same that is used to
        # create new bookmarks, we're just specifying the callback to actually
        # change an existing one instead of create something new in the tree.
        e = Dialogs.NewBookmarkDialog(parent_widget,
                                      editReferenceStub, data)
    else:
        # Same as above (for bookmark dialogs) except this is for folder
        # names.
        e = Dialogs.NewFolderDialog(parent_widget,
                                    editReferenceStub, data.getName())
    return None

#------------------------------------------------------------------------------
# default routine to get contents of subtree
# supply this for a different type of app
# argument is the node object being expanded
# should return list of 4-tuples in the form:
# (label, unique identifier, closed icon, open icon)
# where:
#    label             - the name to be displayed
#    unique identifier - an internal fully unique name
#    closed icon       - PhotoImage of closed item
#    open icon         - PhotoImage of open item, or None if not openable
def get_contents(node):
    """Returns the contents of a particular node"""

    def fn(node):
        """Just a function used for traversing the list inside of a
        BookmarkMenu object.  See Bookmark.BookmarkMenu for information on
        how they are put together"""
        node = node.getData()
        if node.__class__ == Bookmark.BookmarkMenu:
            node_name = node.getName()
            tuple = (node_name, node, Icons.SHUT_ICON, Icons.OPEN_ICON)
        else:
            node_name = node.getName()
            tuple = (node_name, node, Icons.FILE_ICON, None)

        return tuple
    
    if node.id.__class__ == Bookmark.BookmarkMenu:
        l = node.id.traverse(fn)
        return l
    else:   # It's a Bookmark.ListItem object
        tuple = (node.id.getData().getName(), node.id.getData(),
                 Icons.FILE_ICON, None)
        return tuple
                
#------------------------------------------------------------------------------
class Tree(Canvas):
    def __init__(self, master, datatree, rootlabel=None,
                 openicon=None, shuticon=None, getcontents=get_contents,
                 init=None, before=None, after=None, cut=None, paste=None,
                 dnd=None, distx=15, disty=15, textoff=10, lineflag=1,
                 **kw_args):
        self.data_tree = datatree
        self.cutbuffer = None
        
        # pass args to superclass
        apply(Canvas.__init__, (self, master), kw_args)
        
        # default images (BASE64-encoded GIF files)
        # we have to delay initialization until Tk starts up or PhotoImage()
        # complains (otherwise I'd just put it up top)
        if Icons.OPEN_ICON == None:
            Icons.OPEN_ICON = PhotoImage(
                data='R0lGODlhEAANAKIAAAAAAMDAwICAgP//////ADAwMAAAAAAA' \
                'ACH5BAEAAAEALAAAAAAQAA0AAAM6GCrM+jCIQamIbw6ybXNSx3GVB' \
                'YRiygnA534Eq5UlO8jUqLYsquuy0+SXap1CxBHr+HoBjoGndDpNAAA7')
            Icons.SHUT_ICON = PhotoImage(
                data='R0lGODlhDwANAKIAAAAAAMDAwICAgP//////ADAwMAAAAAAA' \
                'ACH5BAEAAAEALAAAAAAPAA0AAAMyGCHM+lAMMoeAT9Jtm5NDKI4Wo' \
                'FXcJphhipanq7Kvu8b1dLc5tcuom2foAQQAyKRSmQAAOw==')
            Icons.FILE_ICON = PhotoImage(
                data='R0lGODlhCwAOAJEAAAAAAICAgP///8DAwCH5BAEAAAMALAAA' \
                'AAALAA4AAAIphA+jA+JuVgtUtMQePJlWCgSN9oSTV5lkKQpo2q5W+' \
                'wbzuJrIHgw1WgAAOw==')
            
        # function to return subnodes (not very much use w/o this)
        if not getcontents:
            raise ValueError, 'must have "get_contents" function'
        
        self.get_contents = getcontents
        
        # horizontal distance that subtrees are indented
        self.distx = distx
        
        # vertical distance between rows
        self.disty = disty
        
        # how far to offset text label
        self.textoff = textoff

        # self.item_dbl_click_hook = self.editNode
        self.item_dbl_click_hook = None
        
        # called after new node initialization
        self.init_hook = init
        
        # called just before subtree expand/collapse
        self.before_hook = before
        
        # called just after subtree expand/collapse
        self.after_hook = after
        
        # called at the end of the cut operation
        self.cut_hook = cut
        
        # called beginning of the paste operation
        self.paste_hook = paste
        
        # flag to display lines
        self.lineflag = lineflag
        
        # called at end of drag'n'drop operation
        # self.dnd_hook = dnd
        self.dnd_hook = None
        
        # create root node to get the ball rolling
        if openicon:
            oi = openicon
        else:
            oi = Icons.OPEN_ICON
        if shuticon:
            si = shuticon
        else:
            si = Icons.SHUT_ICON

        if not rootlabel:
            rootlabel = " "

        self.root = Node(parent=None,
                         name=rootlabel,
                         id=self.data_tree,
                         closed_icon=si,
                         open_icon=oi,
                         x=10,
                         y=10,
                         parentwidget=self)
        
        self.root.drawNode()

        # configure for scrollbar(s)
        x1, y1, x2, y2 = self.bbox('all') 
        self.configure(scrollregion=(x1, y1, x2+5, y2+5))

        # add a cursor
        self.cursor_box = self.create_rectangle(0, 0, 0, 0)
        self.move_cursor(self.root)

        # make it easy to point to control
        self.bind('<Enter>', self.mousefocus)

        # totally arbitrary yet hopefully intuitive default keybindings
        # page-up/page-down
        self.bind('<Next>', self.pagedown)
        self.bind('<Prior>', self.pageup)

        # arrow-up/arrow-down
        self.bind('<Down>', self.next)
        self.bind('<Up>', self.prev)

        # arrow-left/arrow-right
        self.bind('<Left>', self.ascend)
        # (hold this down and you expand the entire tree)
        self.bind('<Right>', self.descend)

        # home/end
        self.bind('<Home>', self.first)
        self.bind('<End>', self.last)

        # space bar
        self.bind('<Key-space>', self.toggle)

    def destroy(self, *args):
        # Garbage collect the image objects.  If this isn't done,
        # Tkinter holds onto a useless reference to them which can cause
        # a stderr message.
        Icons.OPEN_ICON = None
        Icons.SHUT_ICON = None
        Icons.FILE_ICON = None

        # Superclass destructor
        Canvas.destroy(self)
        return None

    def editNode(self, node):
        pass

    # scroll (in a series of nudges) so items are visible
    def see(self, *items):
        x1, y1, x2, y2 = apply(self.bbox, items)
        while x2 > self.canvasx(0)+self.winfo_width():
            old = self.canvasx(0)
            self.xview('scroll', 1, 'units')
            # avoid endless loop if we can't scroll
            if old == self.canvasx(0):
                break
        while y2 > self.canvasy(0)+self.winfo_height():
            old = self.canvasy(0)
            self.yview('scroll', 1, 'units')
            if old == self.canvasy(0):
                break
        # done in this order to ensure upper-left of object is visible
        while x1 < self.canvasx(0):
            old = self.canvasx(0)
            self.xview('scroll', -1, 'units')
            if old == self.canvasx(0):
                break
        while y1 < self.canvasy(0):
            old = self.canvasy(0)
            self.yview('scroll', -1, 'units')
            if old == self.canvasy(0):
                break
            
    # move cursor to node
    def move_cursor(self, node):
        self.pos = node
        x1, y1, x2, y2 = self.bbox(node.symbol, node.label)
        self.coords(self.cursor_box, x1-1, y1-1, x2+1, y2+1)
        self.see(node.symbol, node.label)
    
    # expand given path
    # note that the convention used in this program to identify a
    # particular node is to give a tuple listing it's id and parent ids
    def expand(self, path):
        self.root.expand(path[1:])

    def setCutBuffer(self, newcutbuffer):
        self.cutbuffer = newcutbuffer
        return self.cutbuffer
    
    def getCutBuffer(self):
        return self.cutbuffer

    def getRoot(self):
        return self.root

    def cut(self):
        """Tree.cut: cut the selected node"""
        active = self.getActive()

        if active != self.root:     # Don't allow the user to cut the root node
            self.cutbuffer = self.getActive().cut()
        return None

    def expandAll(self):
        return self.root.expandAll()

    def paste(self):
        if self.cutbuffer:
            if self.getActive().isFolder() and self.cutbuffer.node.isFolder():
                # If we're pasting a folder into a folder, make it the
                # first item in the list
                self.getActive().paste(self.cutbuffer, 1)
            elif self.getActive().isFolder():
                # If we're pasting a non-folder into a folder, make it the
                # last sibling
                self.getActive().paste(self.cutbuffer, 2)
            else:
                # Otherwise, add the new item as the next sibling of whatever
                # was active.
                self.getActive().paste(self.cutbuffer, 3)
        return None

    def setActive(self, newactive):
        self.pos = newactive
        return self.pos

    def getActive(self):
        return self.pos

    def getTree(self):
        return self.root

    # soak up event argument when moused-over
    # could've used lambda but didn't...
    def mousefocus(self, event):
        self.focus_set()
        
    # open/close subtree
    def toggle(self, event=None):
        self.pos.toggle_state()

    # move to next lower visible node
    def next(self, event=None):
        self.move_cursor(self.pos.next())
            
    # move to next higher visible node
    def prev(self, event=None):
        self.move_cursor(self.pos.prev())

    # move to immediate parent
    def ascend(self, event=None):
        if self.pos.parent:
            # move to parent
            self.move_cursor(self.pos.parent)

    # move right, expanding as we go
    def descend(self, event=None):
        self.pos.toggle_state(1)
        if self.pos.subnodes:
            # move to first subnode
            self.move_cursor(self.pos.subnodes[0])
        else:
            # if no subnodes, move to next sibling
            self.next()

    # go to root
    def first(self, event=None):
        # move to root node
        self.move_cursor(self.root)

    # go to last visible node
    def last(self, event=None):
        # move to bottom-most node
        n = self.root
        while n.subnodes:
            n = n.subnodes[-1]
        self.move_cursor(n)

    # previous page
    def pageup(self, event=None):
        n = self.pos
        j = self.winfo_height()/self.disty
        for i in range(j-3):
            n = n.prev()
        self.yview('scroll', -1, 'pages')
        self.move_cursor(n)

    # next page
    def pagedown(self, event=None):
        n = self.pos
        j = self.winfo_height()/self.disty
        for i in range(j-3):
            n = n.next()
        self.yview('scroll', 1, 'pages')
        self.move_cursor(n)

    # drag'n'drop support using Tkdnd
    # start drag'n'drop
    def dnd_click(self, event, node):
        Tkdnd.dnd_start(self, event)
        self.dnd_source = node

    # remember node we just entered, and save it as target
    def enter(self, node):
        self.dnd_target = node

    # we're not over a valid target
    def leave(self, event):
        self.dnd_target = None
        
    # end drag'n'drop
    def dnd_end(self, target, event):
        pass
    
    def dnd_accept(self, source, event):
        return self

    def dnd_commit(self, source, event):
        # destroy the move icon
        self.dnd_leave(None, None)
        # force update to get <Enter> event, if any
        self.update()
        # see if non-trivial drag'n'drop occurred
        if self.dnd_target == None or source.dnd_source == self.dnd_target:
            return
        self.dnd_hook(source, self.dnd_target)

    # create drag icon
    def dnd_enter(self, source, event):
        # returns pointer position in display coordinates
        x, y = self.winfo_pointerxy()
        # translate to canvas coordinates
        x = self.canvasx(x)-self.winfo_rootx()
        y = self.canvasy(y)-self.winfo_rooty()
        i = source.itemcget(source.dnd_source.symbol, 'image')
        self.dnd_symbol = self.create_image(x, y, image=i)
        i = source.itemcget(source.dnd_source.label, 'text')
        self.dnd_label = self.create_text(x+self.textoff, y, 
                                          text=i, justify='left',
                                          anchor='w' )
                                             
    # destroy drag icon
    def dnd_leave(self, source, event):
        self.delete(self.dnd_symbol, self.dnd_label)

    # move drag icon
    def dnd_motion(self, source, event):
        # returns pointer position in display coordinates
        x, y = self.winfo_pointerxy()
        # translate to canvas coordinates
        x = self.canvasx(x)-self.winfo_rootx()
        y = self.canvasy(y)-self.winfo_rooty()
        self.coords(self.dnd_symbol, x, y)
        self.coords(self.dnd_label, x+self.textoff, y)



# GopherResponse.py
# $Id: GopherResponse.py,v 1.15 2001/07/09 22:32:14 s2mdalle Exp $
# Contains GopherResource and GopherResponse objects
# Written by David Allen <mda@idatar.com>
# Released under the terms of the GNU General Public License
#
# This object holds the data corresponding to how a gopher server responded
# to a given request.  It holds one of two things in general, either a pile
# of data corresponding to text, a gif file, whatever, or it holds a list of
# GopherResource objects.  (This is when it's a directory entry that's being
# stored.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##############################################################################
import re
from string              import *
from urlparse            import *
from gopher              import *
import Connection
import GopherConnection
import GopherObject
import GopherResource
import ResourceInformation
import Options

GopherResponseException = "Gopher response error"

class GopherResponse(GopherObject.GopherObject):
    verbose = None
    def __init__(self, type=None, host=None, port=None, loc=None, name=None):
        GopherObject.GopherObject.__init__(self, type, host, port, loc, name)
        self.__class = "GopherResponse"
        self.data = None
        self.responses = []
        return None
    def toProtocolString(self):
        if self.getData() == None:
            def protString(item):
                return item.toProtocolString()
            return join(map(protString, self.getResponses()), "")
        else:
            return self.getData()
        return None
    def writeToFile(self, filename):
        """Writes the contents of this response to a disk file at filename
        this may raise IOError which the caller should deal with"""
        fp = open(filename, "w")
        
        if self.getData() == None:
            for resp in self.getResponses():
                fp.write(resp.toProtocolString())
        else:
            fp.write(self.getData())

        fp.flush()
        fp.close()
        return filename
    def getResponses(self):
        """Return a list of responses.  This is only really good when the
        response was a directory and set of entries."""
        if self.getData() != None:
            raise GopherResponseException, "Get data instead."
        return self.responses
    def getData(self):
        """Return the data associated with the response.  This is usually all
        of the data off of the socket except the trailing closer."""
        return self.data
    def getDataLength(self):
        return len(self.data)
    def getDataChunk(self, startIndex, endIndex=-1):
        """This method provides a way to get chunks of data at a time,
        rather than snarfing the entire ball."""
        
        if endIndex == -1:
            endIndex = len(self.data)
        return self.data[startindex, endIndex]
    def getError(self):
        """If there was an error message, return it."""
        try:
            return self.error
        except:
            return None
    def setError(self, errorstr):
        """Modify the error message within this response."""
        self.error = errorstr
        return self.getError()
    def setData(self, data):
        """Modify the data within the response.  You probably don't want to
        do this."""
        self.data = data
        return None
    def looksLikeDir(self, data):
        """Takes a chunk of data and returns true if it looks like directory
        data, and false otherwise.  This is tricky, and is of course not 100%.
        Basically, we look at each line, see if the first bit in the line is a
        legal type, check to see that the line has at least 3 tabs in it.  If
        all of the first 20 lines of the data follow that rule, then it's good
        enough to be used as directory data.  If not, it gets chucked.  Notice
        that if this really is a directory but it's using file types we've
        never heard of (see gopher.py) then it will still get thrown out.
        Bummer.  This should only be called anyway if the type indictator is
        incorrect, so cope.  :)"""
        
        def linefn(l):
            return replace(l, "\r", "")

        # Some very strange non-standards compliant servers send \r on some
        # lines and not on others.  So split by newline and remove all
        # carriage returns as they occur.
        lines = map(linefn, split(data, "\n", 10))

        for line in lines:
            d = strip(line)
            if not d or d == '' or d == '.':
                continue
            
            if count(line, "\t") < 2:
                return None               # Not enough tabs.  Bummer.

            isResponse = None
            byte = line[0]
            try:
                resp = responses[byte]
                isRespose = 1
            except:
                pass

            if isResponse:
                continue
            
            try:
                resp = errors[byte]
            except:
                return None

        if len(lines) > 0:
            return 1     # Matched all tests for max 20 lines.  Looks OK.
        else:
            return 0     # Empty data isn't a directory.
    
    def parseResponse(self, data):
        """Takes a lump of data, and tries to parse it as if it was a directory
        response.  It will set the responses array to the proper thing if the
        result was good, so that you can use self.getRepsonses() to access
        them.  Otherwise it raises GopherResponseException"""
        
        self.responses = []

        if self.type == RESPONSE_DIR:
            pass          # Keep going
        elif self.looksLikeDir(data):
            self.type = RESPONSE_DIR
        else:
            raise GopherException, "This isn't a directory."

        def stripCR(dat):
            return replace(dat, "\r", "")

        # This is done because \r may or may not be present, so we can't
        # split by \r\n because it fails for some misbehaved gopher servers.
        self.lines = map(stripCR, split(data, "\n"))

        for line in self.lines:
            if len(line) <= 1:
                continue

            # Type of the resource.  See gopher.py for the possibilities.
            stype = "%s" % line[0]
            line = line[1:]

            # Gopher protocol uses tab delimited fields...
            linedata = split(line, "\t")
            name    = "Unknown"              # Silly defaults
            locator = "Unknown"
            host    = "Unknown"
            port    = 70

            try:
                name = linedata[0]           # Assign the right things in the
            except IndexError: pass          # right places (maybe)
            try:                             # Catch those errors coming from
                locator = linedata[1]        # the line not being split into 
            except IndexError: pass          # enough tokens.  Realistically,
            try:                             # if those errors happen,
                host = linedata[2]           # something is really screwed up
            except IndexError: pass          # and there's not much we can do
            try:                             # anyway.  
                port = linedata[3]
            except IndexError: pass

            try:
                remainder = linedata[4:]     # Extra Gopher+ fields.
            except:
                remainder = []

            # UMN gopherd errors do this sometimes.  It's quite annoying.
            # they list the response type as 'directory' and then put the host
            # as 'error.host' to flag errors
            if host == 'error.host' and stype != RESPONSE_BLURB:
                stype = RESPONSE_ERR

            newresource = GopherResource.GopherResource(stype, host,
                                                        port, locator, name,
                                                        remainder)
            
            # Are the options set to allow us to get info?
            # THIS SHOULD BE USED WITH CAUTION since it can slow things down
            # more than you might think.
            if Options.program_options.getOption('grab_resource_info'):
                if len(remainder) >= 1 and remainder[0] == '+':
                    try:
                        conn = GopherConnection.GopherConnection()
                        info = conn.getInfo(newresource)
                        newresource.setInfo(info)
                    except GopherConnectionException, estr:
                        print "***(GCE) can't get info: %s" % estr
                    except Exception, estr:
                        print "***(unknown) Couldn't %s %s" % (Exception,estr)

            self.responses.append(newresource)
        return None
# End GopherResponse



#!/usr/bin/python
# FORG.py
# $Id: forg.py,v 1.15 2001/09/02 17:01:42 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
#
# This file contains the core FORG class, which handles most of the logic
# and content.  The actual GUI around it is handled by TkGui.py.
# Theoretically, you could replace TkGui with GtkGui and it would work, but
# not really, since you'd have to rewrite the Tkinter specific parts of
# GUI*.py
#
# This object is meant as a wrapper for other things that subclass ContentFrame
# This should simply be the object that is packed into the main GUI window
# and be a subclass of Tkinter.Frame.
#
# This object knows how to download resources and display them properly by
# using the other classes in the FORG.  It basically is just a shell of a
# Tkinter.Frame that contains the logic for using the right classes in the
# right way.  So without navigation, this contains all that is needed to
# implement the functionality in other programs using just this class (and of
# course the other things that it needs)
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##############################################################################
# System wide imports
from Tkinter   import *                # Tk interface
from string    import *                # String manipulation
from threading import *                # Threads
import Pmw                             # Python Mega Widgets
import os                              # Operating system stuff
import socket                          # Socket communication
import sys
import tkFileDialog

# Non-GUI FORG specific imports
from gopher import *
import utils
import GopherConnection
import AskForm
import GopherResource
import GopherResponse
import Cache
import Options
import Associations
import List
import ListNode
import State

# GUI Specific
import GUIAskForm
import GUIDirectory
import GUIError
import GUIFile
import GUIQuestion
import GUISaveFile
import GUISearch
import ResourceInformation
import Dialogs

#--------------------------- VERSION ------------------------------------
VERSION = '0.5.1'
#--------------------------- VERSION ------------------------------------

FORGException = "Some error came up. I sure hope it came with a real message"
FORGContentException = "Content was missing probably"

def getVersion():
    return VERSION

class FORG(Frame):
    verbose = None
    
    def __init__(self, parent_widget, resource, response=None,
                 messageBar=None,
                 parent_object=None):
        Frame.__init__(self, parent_widget)
        self.parent_widget = parent_widget
        self.parent_object = parent_object
        self.response      = response
        self.resource      = resource
        self.mb            = messageBar
        self.opts          = Options.program_options  # Global alias

        self.locations      = List.List()
        self.navList        = List.List()

        self._createPopup()
        
        self.currentContent = None

        # Go wherever the user intended us to go.
        self.goElsewhere(resource)
        
        return None
    def getCache(self):
        """Return the cache object being used"""
        return Options.program_options.getCache()
    def getMessageBar(self):
        """Return the message bar being used."""
        return self.mb
    def setMessageBar(self, newmb):
        """Set the message bar object."""
        self.mb = newmb
        return self.getMessageBar()

    def dumpQueue(self, *args):
        print "QUEUE IS:"
        def fn(node, s=self):
            r = node.getData().getResource()
            print r.toURL()
            
        self.navList.traverse(fn)

    def stats(self, *args):
        buffer = "%s item(s) in the queue\n" % self.navList.countNodes()
        buffer = buffer + Options.program_options.cache.getCacheStats() + "\n"

        if self.verbose:
            print "QUEUE IS:"
            def fn(node, s=self):
                r = node.getData().getResource()
                print r.toURL()
            
        self.navList.traverse(fn)
            
        self.genericMessage(buffer)
    
    def find(self, term, caseSensitive=None, lastIdentifier=None):
        """This find method forwards all arguments to the child widget and
        does no processing.  See the documentation for ContentFrame.find for
        information on what parameters mean and what the method does."""
        # Pass on to the child verbatim.
        return self.child.find(term, caseSensitive, lastIdentifier)

    def genericMessage(self, str, title='Information:'):
        """Given a string, pops up a dialog with the appropriate title and
        content string."""
        x = Dialogs.InformationDialog(self, str, title)
        return None

    def genericError(self, errstr, title='Error:'):
        """Given an error string, pops up a dialog with the same title and
        text, alerting the user to something or other."""
        x = Dialogs.ErrorDialog(self, errstr, title)
        return None
    
    def getResource(self):
        """Returns the stored GopherResource object that was used to create
        this object."""
        return self.resource
    
    def getResponse(self):
        """Returns the stored GopherResponse object corresponding to the
        resource passed during the creation of this object"""
        return self.response

    def getGopherResponse(self):
        """This fetches the resource handed to the object from the network
        if necessary."""
        self.conn = GopherConnection.GopherConnection()
        msg = "Done"
        try:
            self.response = self.conn.getResource(resource=self.resource,
                                                  msgBar=self.mb)
        except GopherConnection.GopherConnectionException, estr:
            raise FORGException, estr
        except socket.error, err:
            try:
                if len(err) >= 2:
                    msg = "Can't fetch: errcode %d: %s" % (err[0], err[1])
                else:
                    msg = "Can't fetch: error %s" % err[0]
            except AttributeError:  # This is really weird, don't ask.
                msg =  "Can't fetch - unknown error."
            err = "getGopherResponse: %s" % err
            raise FORGException, err

        if self.response.getError():
            raise FORGException, self.response.getError()

        # Turn the green light back on, since we're done with the transmission.
        self.opts.greenLight()
        return None

    def popupMenu(self, event):
        """Display pop-up menu on right click on a message"""
        self.popup.tk_popup(event.x_root, event.y_root)

    def saveFile(self, *args):
        """Pop up a dialog box telling the user to choose a filename to save
        the file as."""
        dir = os.path.abspath(os.getcwd())
        filename = tkFileDialog.asksaveasfilename(initialdir=dir)
        
        listNode = self.navList.getCurrent()
        response = listNode.getData().getResponse()

        if filename and filename != "":
            try:
                response.writeToFile(filename)
            except IOError, errstr:
                self.genericError("Couldn't write data to\n%s:\n%s" %
                                  (filename, errstr))

        return None

    def _createPopup(self):
        """Creates Pop-up menu on right click on a message"""
        self.popup = Menu(self)
        self.popup['tearoff'] = FALSE
        self.popup.add_command(label='Save', command=self.saveFile)
        self.popup.add_command(label='Back', command=self.goBackward)
        self.popup.add_command(label='Forward', command=self.goForward)

    def goForward(self, *rest):
        """Navigate forward"""
        try:
            node = self.navList.getNext()
            data = node.getData()

            # In case we are going forward to an error message...
            if not data.getResponse():
                # Remove current item in the list.
                self.navList.removeCurrent()

                # Go to this location again.
                return self.goElsewhere(data.getResource())

            self.createResponseWidget()
            self.changeContent(self.child)
        except List.ListException, errstr:
            if self.verbose:
                print "Couldn't get next: %s" % errstr
        return None

    def goBackward(self, *rest):
        """Navigate backward."""
        try:
            node = self.navList.getPrev()
            state = node.getData()

            # In case we are going back to an error message...
            if not state.getResponse():
                # Remove current item in the list.
                self.navList.removeCurrent()

                # Go to this location again.
                return self.goElsewhere(state.getResource())

            self.response = state.getResponse()
            self.resource = state.getResource()
            self.createResponseWidget()
            node.setData(State.State(self.response, self.resource, self.child))
            self.changeContent(self.child)
        except List.ListException, errstr:
            if self.verbose:
                print "Couldn't get prev: %s" % errstr

    def downloadResource(self, res):
        """Downloads the resource from the network and redisplays"""
        self.resource = res

        try:
            self.getGopherResponse()
        except FORGException, estr:
            self.genericError("Error fetching resource:\n%s" % estr)
            return None
            
        self.createResponseWidget()
        self.navList.insert(ListNode.ListNode(State.State(self.response,
                                                          self.resource,
                                                          self.child)))
        self.changeContent(self.child)
        return None

    def reload(self):
        newthread = Thread(group=None,
                           target=self.reloadResource,
                           name="Reload Thread",
                           args=())
        # Thread is responsible for adding the resource to the queue list
        # It will run and do its thing and when the function returns, the \
        # thread ends.
        newthread.start()
        return None

    def reloadResource(self):
        """Reloads from the network the current resource."""
        try:
            self.getGopherResponse()
        except FORGException, estr:
            self.genericError("Error fetching resource:\n%s" % estr)
            return None
        
        self.createResponseWidget()
        s = State.State(self.response, self.resource, self.child)
        self.navList.getCurrent().setData(s)
        self.changeContent(self.child)
        return None

    def getCurrentURL(self):
        return self.resource.toURL()
    
    def goElsewhere(self, resource, usecache=1, *args):
        """This is a wrapper for dealing with the download thread.  Pass it
        one resource as an argument, and that resource will be downloaded in
        the background.  Optionally, if the file exists in the cache, it will
        be loaded from there if usecache is true."""

        self.resource   = resource
        self.response   = None
        optionsUseCache = Options.program_options.getOption("use_cache")
        
        # Attempt to load the response object out of the cache.
        if usecache and optionsUseCache and not resource.getDataBlock():
            # Don't try to pull something out of cache unless it doesn't have
            # a data block. If it does, we have to submit information
            utils.msg(self.mb, "Looking for document in cache...")
            self.response = Options.program_options.cache.uncache(resource)

        if not self.response:
            utils.msg(self.mb,
                      "Document not in cache. Fetching from network.")
            # We need to fetch this document from the network.
            # Signal the download thread to go to work, and get out.
            newthread = Thread(group=None,
                               target=self.downloadResource,
                               name="Download Thread",
                               args=(self.resource,))
            # Thread is responsible for adding the resource to the queue list
            newthread.start()
            return
        else:
            utils.msg(self.mb, "Loading document from cache to display.")
            self.createResponseWidget()
            s = State.State(self.response, self.resource, self.child)
            self.navList.insert(ListNode.ListNode(s))
            self.changeContent(self.child)
        return None

    def changeContent(self, newwid):
        """Changes the current main content of the window to newwid"""
        if self.currentContent:
            self.currentContent.pack_forget()

        if newwid:
            utils.msg(self.mb, "Updating display...")            
            newwid.pack(expand=1, fill='both')
            newwid.pack_content()
            utils.msg(self.mb, "Done")

        self.currentContent = newwid

        try:
            self.parent_object.change_content_hook(self.getResource())
        except:
            # If it doesn't exist, or fails for some reason, that's the
            # parents problem, not ours.
            pass
        
        return self.currentContent
    
    def createResponseWidget(self):
        """Take the response already fetched and turn it into a child widget.
        This only works if the resource has already been fetched from the
        network."""
        
        if not self.response:
            raise(FORGException,
                  "createResponseWidget: No valid response present")
        if not self.resource:
            raise(FORGException,
                  "createResponseWidget: No valid resource present")

        cfilename = None
        
        if self.opts.getOption('use_cache'):
            utils.msg(self.mb, 'Caching data...')
            cfilename = ''

            _resr = self.resource.shouldCache()
            _resp = self.response.shouldCache()
            
            if not self.resource.isAskType() and _resr and _resp:
                # Don't try to cache ASK blocks.  It will only throw an
                # exception since caching them isn't a very stellar idea.
                try:
                    cfilename = self.opts.cache.cache(resp=self.response,
                                                      resource=self.resource)
                except Cache.CacheException, exceptionstr:
                    self.genericError(exceptionstr)

        r = self.response

        if self.response.getTypeCode() == RESPONSE_INDEXS:
            self.child = GUISearch.GUISearch(parent_widget=self,
                                             parent_object=self,
                                             resp=self.response,
                                             resource=self.resource,
                                             filename=cfilename)
        elif self.response.__class__ == AskForm.AskForm:
            try:
                self.child = GUIAskForm.GUIAskForm(parent_widget=self,
                                                   parent_object=self,
                                                   resp=self.response,
                                                   resource=self.resource,
                                                   filename=cfilename)
            except Exception, errstr:
                print "******************************************************"
                print "Caught ", Exception, " with error ", errstr, " ASK"
                self.child = Label(self, text="Congrats!  You've found a bug!")
        elif r.getData() != None and not r.looksLikeDir(r.getData()):
            if cfilename != None:
                assoc = self.opts.associations.getAssociation(cfilename)
                
                if assoc != None:
                    self.LAUNCH_ME = [cfilename, assoc]
                    
            if self.response.getTypeCode() == RESPONSE_FILE:
                self.child = GUIFile.GUIFile(parent_widget=self,
                                             parent_object=self,
                                             resp=self.response,
                                             resource=self.resource,
                                             filename=cfilename)
            else:
                self.child = GUISaveFile.GUISaveFile(parent_widget=self,
                                                     parent_object=self,
                                                     resp=self.response,
                                                     resource=self.resource,
                                                     filename=cfilename)

        else:  # There is no data, display a directory entry
            ma = {
                "Back"       : self.goBackward,
                "Forward"    : self.goForward,
                "About FORG" : self.about }
                
            self.child = GUIDirectory.GUIDirectory(parent_widget=self,
                                                   parent_object=self, 
                                                   resp=self.response,
                                                   resource=self.resource,
                                                   filename=cfilename,
                                                   menuAssocs=ma)

    def about(self, *args):
        """Display the about box."""
        Pmw.aboutversion(getVersion())
        Pmw.aboutcopyright('Copyright 2000, 2001')
        Pmw.aboutcontact(
            'This program is licensed under the GNU General Public License\n' +
            'For more information, please see\n' +
            'http://www.gnu.org/')
        self._about_dialog = Pmw.AboutDialog(self, applicationname='FORG')
        # self._about_dialog.show()
        return None


################################### MAIN CODE ###############################

if __name__ == '__main__':
    # Require TkGui only if we are running the application stand-alone.
    from TkGui import *

    print "Starting the FORG"

    arr = sys.argv
    
    def main(item):
        try:
            app = TkGui(item)
            app.mainloop()
        except KeyboardInterrupt:
            app.quit()
            exit()

    url = ""

    try:
        url = arr[1]
    except:
        url = "gopher://gopher.floodgap.com:70/"

    if find(lower(url[:]), "gopher://") == -1:
        url = "gopher://" + url

    # Start the program
    print "Starting program with \"%s\"" % url
    main(url)




# GUISearch.py
# $Id: GUISearch.py,v 1.6 2001/04/15 19:27:00 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
#
# Released under the terms of the GNU General Public License
#
# This is the graphical component used for getting information from the user
# about search terms.  (And then sending them on their merry way)
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##############################################################################

from Tkinter import *
from gopher import *
import Pmw
import ContentFrame
import GopherResource
import GopherResponse
import GopherConnection

class GUISearch(ContentFrame.ContentFrame, Frame):
    verbose = None
    
    def __init__(self, parent_widget, parent_object, resp,
                 resource, filename=None, menuAssocs={}):
        Frame.__init__(self, parent_widget)  # Superclass constructor
        self.resource = resource
        self.response = resp
        self.parent   = parent_object
        self.filename = filename

        labels = ["Please enter a space-separated list of search terms."]

        last = 0
        
        for x in range(0, len(labels)):
            last = x
            label = labels[x]
            Label(self, text=label, foreground='#000000').grid(row=x, col=0,
                                                               columnspan=2)

        self.entryArea = Frame(self)
        self.entryArea.grid(row=(x+1), col=0, columnspan=5, sticky='EW')
        self.entryBox = Entry(self.entryArea, text='')
        self.entryBox.pack(side='left', expand=1, fill='x')
        self.entryBox.bind("<Return>", self.submit)
        self.GO = Button(self.entryArea, text='Submit', command=self.submit)
        self.GO.pack(side='right')
        self.bottom_label = None
        return None
    def pack_content(self, *args):
        return None
    def find(self, term, caseSensitive=None, lastIdentifier=None):
        self.parent.genericError("Error:  Search boxes\n" +
                                 "are not searchable.")
        return None
    def submit(self, *args):
        terms = self.entryBox.get()

        print "Terms are \"%s\"" % terms

        if self.bottom_label:
            self.bottom_label.destroy()
            self.bottom_label = Label(self, "Searching for \"%s\"" % terms)
            self.bottom_label.grid(row=10, col=0, columnspan=2,
                                   sticky=W)

        # Copy the data from the current resource.
        res = GopherResource.GopherResource()
        res.setHost(self.resource.getHost())
        res.setPort(self.resource.getPort())
        res.setName(self.resource.getName())

        # This is a nasty way of faking the appropriate protocol message,
        # but oh well...
        res.setLocator("%s\t%s" % (self.resource.getLocator(),
                                   terms))
        res.setInfo(None)
        res.setLen(-2)
        res.setTypeCode(RESPONSE_DIR)   # Response *will* be a directory.

        self.parent.goElsewhere(res)
        return None
    




# Dialogs.py
# $Id: Dialogs.py,v 1.5 2001/09/01 17:39:39 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
#
# Contains many different program dialogs used for information and data
# entry purposes.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
############################################################################
from Tkinter import *
from types import *
from string import *
import Pmw

import GopherResource
import Bookmark

class FindDialog:
    def __init__(self, parent, searchwidget, parentobj=None, *args):
        self.parent = parent
        self.parentobj = parentobj
        self.searchwidget = searchwidget
        self.dialog = Pmw.Dialog(parent, title='Find Text...',
                                 buttons=('OK', 'Cancel'), defaultbutton='OK',
                                 command=self.dispatch)
        self.frame = Frame(self.dialog.interior())
        self.frame.pack(expand=1, fill='both')

        Label(self.frame, text="Find a term...").grid(row=0, col=0,
                                                      columnspan=5)
        Label(self.frame, text="Term: ").grid(row=1, col=0)

        self.searchEntry = Entry(self.frame, text="")
        self.searchEntry.grid(row=1, col=1, columnspan=4)

        self.css = IntVar()
        
        self.caseSensitiveCheckBox = Checkbutton(self.frame,
                                                 text="Case-sensitive search",
                                                 variable = self.css,
                                                 command  = self.cb)
        self.caseSensitiveCheckBox.grid(row=2, col=0, columnspan=4)
        
        self.lastMatch = None
        # self.dialog.activate()
        return None
    
    def cb(self, *args):
        # print "Var is ", self.css.get()
        return None
    
    def getSearchTerm(self):
        """Returns the search term currently in the search box."""
        return self.searchEntry.get()
    
    def getCaseSensitive(self):
        """Returns the status of the case sensitive check button"""
        return self.css.get()
    
    def dispatch(self, button):
        """Handles button clicking in the dialog.  (OK/Cancel)"""
        if button != 'OK':
            # Smack...
            self.dialog.destroy()
            return None

        # Otherwise, go look for a term...
        # try:
        self.lastMatch = self.searchwidget.find(self.getSearchTerm(),
                                                self.getCaseSensitive(),
                                                self.lastMatch)
        
        print "Last match is now ", self.lastMatch
        return self.lastMatch

class OpenURLDialog:
    def __init__(self, parentwin, callback):
        self.callback = callback
        self.dialog   = Pmw.Dialog(parentwin, title="Open URL:",
                                   command=self.dispatch,
                                   buttons=('OK', 'Cancel'),
                                   defaultbutton='OK')
        i = self.dialog.interior()
        Label(i, text="Enter URL to Open:").pack(side='top', expand=1,
                                                 fill='both')
        self.urlEntry = Entry(i, width=30)
        self.urlEntry.insert('end', "gopher://")
        self.urlEntry.pack()
        
    def dispatch(self, button):
        if button == 'OK':
            # If OK is clicked, fire the callback with whatever the URL
            # happens to be.
            self.callback(self.urlEntry.get())

        # In any case, destroy the dialog when finished.
        self.dialog.destroy()
        return None
    

class NewBookmarkDialog:
    def __init__(self, parentwin, cmd, resource=None):
        self.cmd = cmd
        self.dialog = Pmw.Dialog(parentwin, title="New Bookmark:",
                                 command=self.callback,
                                 buttons=('OK', 'Cancel'))
        i = self.dialog.interior()
        namebox = Frame(i)
        urlbox  = Frame(i)
        Label(i, text="Enter Bookmark Information:").pack(side='top', expand=1,
                                                          fill='both')
        namebox.pack(fill='both', expand=1)
        urlbox.pack(fill='both', expand=1)
        
        Label(namebox, text="Name:").pack(side='left')
        self.nameEntry = Entry(namebox, width=30)
        self.nameEntry.pack(side='right', fill='x', expand=1)
        Label(urlbox, text="URL:").pack(side='left')
        self.URLEntry = Entry(urlbox, width=30)
        self.URLEntry.pack(side='right', fill='x', expand=1)

        if resource:
            self.URLEntry.insert('end', resource.toURL())
            self.nameEntry.insert('end', resource.getName())

        return None

    def callback(self, button):
        if button and button != 'Cancel':
            res = Bookmark.Bookmark()
            res.setURL(self.URLEntry.get())            
            res.setName(self.nameEntry.get())
            self.cmd(res) # Do whatever our parent wants us to with this...

        self.dialog.destroy()
        return None

class NewFolderDialog:
    BUTTON_OK     = 0
    BUTTON_CANCEL = 1
    
    def __init__(self, parentwin, callback, folderName=None):
        self.buttons = ('OK', 'Cancel')
        self.callback = callback
        self.parent   = parentwin
        self.dialog = Pmw.Dialog(parentwin,
                                 title="New Folder:",
                                 command=self.closeDialog,
                                 buttons=self.buttons)

        i = self.dialog.interior()
        Label(i, text="New Folder Title:").grid(row=0, col=0,
                                               sticky='EW')
        self.__entry = Entry(i)        
        self.__entry.grid(row=1, col=0, sticky='EW')
        
        if folderName:
            self.__entry.insert('end', folderName)
            
        self.bindCallbacks()
        return None

    def bindCallbacks(self):
        self.__entry.bind('<Return>', self.closeDialog)
        return None
    
    def closeDialog(self, result):
        if result == self.buttons[self.BUTTON_CANCEL]:
            self.dialog.destroy()
            return None
        else:
            str = self.__entry.get()
            self.dialog.destroy()
            return self.callback(str)
    # End NewFolderDialog
    
class InformationDialog:
    def __init__(self, parent, errstr, title='Information:'):
        # We don't need an activate command since we want the dialog to just
        # get smacked when the user presses close.
        if title == '':
            title = errstr
        self.dialog = Pmw.Dialog(parent, title=title, buttons=["Close"])

        # print "========================================================"
        # print "Error Dialog: %s" % errstr
        # print "========================================================"

        if type(errstr) != StringType:
            errstr = errstr.__str__

        labels = split(errstr, "\n")
        
        for label in labels:
            Label(self.dialog.interior(), text=label).pack(side='top')

        # self.dialog.activate() # Modalize  :)
        return None

class ErrorDialog(InformationDialog):
    def __init__(self, parent, errstr, title="Error:"):
        InformationDialog.__init__(self, parent, errstr, title)
        return None






# Question.py
# Written by David Allen <mda@idatar.com>
# Released under the terms of the GNU General Public License
# $Id: Question.py,v 1.6 2001/07/11 22:43:09 s2mdalle Exp $
# Represents one question inside of a multi-question ASK block
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
############################################################################

from gopher import *
from string import *

QuestionException = "Error dealing with Question"

class Question:
    RESPONSE_ONE_LINE   = 1
    RESPONSE_PASSWORD   = 2
    RESPONSE_MULTI_LINE = 3
    RESPONSE_FILENAME   = 4
    RESPONSE_CHOICES    = 5
    verbose = None
    
    def __init__(self, data=""):
        self.qtype = None
        self.promptString = "%s%s" % ("Answer the nice man's question,\n",
                                      "and nobody gets hurt.")
        self.default      = ""
        self.setData(data)
        return None
    
    def getDefault(self):
        return self.default
    
    def setDefault(self, newdefault):
        self.default = newdefault
        return self.default
    
    def getPromptString(self):
        return self.promptString
    
    def getType(self):
        if self.verbose:
            print "QTYPE is ", self.qtype
        return self.qtype
    
    def setData(self, data):
        """Given data on a question in ASK format, this parses the data
        and sets the internal data of the object correctly.  This should be
        done pretty much first after creating the object."""
        
        self.linedata = data[:]   # Copy

        ind = find(data, ":")
        if ind == -1:
            raise QuestionException, "Cannot find \":\" on line"
        qtype = strip(data[0:ind])
        data = data[ind+1:]

        try:
            self.qtype = questions[qtype]
        except KeyError:
            raise(QuestionException,
                  "Question type \"%s\" is not valid" % qtype)
        
        # Do the rest here...
        if (self.qtype    == QUESTION_ASK
            or self.qtype == QUESTION_ASKL
            or self.qtype == QUESTION_ASKP):
            if find(data, "\t") != -1:
                try:
                    [promptStr, default_val] = split(data, "\t")
                except:
                    raise QuestionException, "Too many tabs in line"
                self.promptString = strip(promptStr)
                self.default = default_val
                if self.verbose:
                    print "Block has default of ", self.default
            else:
                self.promptString = strip(data)
        elif self.qtype == QUESTION_ASKP:
            pass
        elif (self.qtype    == QUESTION_ASKL
              or self.qtype == QUESTION_ASKF
              or self.qtype == QUESTION_CHOOSEF
              or self.qtype == QUESTION_NOTE):
            self.promptString = strip(data)
        elif self.qtype == QUESTION_CHOOSE or self.qtype == QUESTION_SELECT:
            try:
                ind = find(data, "\t")
                prompt = data[0:ind]
                opts = split(data[ind+1:], "\t")
            except :
                raise QuestionException, "Too many tabs in line"

            self.promptString = strip(prompt)
            self.options = opts
            self.default = self.options[0]
        else:
            raise QuestionException, "Unknown QType on parse"

        if self.verbose:
            print "Successfully parsed data line: ", self.linedata

        return None
